@@ -95,6 +95,24 @@ transform_Y (double y)
   return page_height - (y - model_ymin);
 }
 
+static bool
+isnan_2BD (BITCODE_2BD pt)
+{
+  return isnan (pt.x) || isnan (pt.y);
+}
+
+static bool
+isnan_2pt (dwg_point_2d pt)
+{
+  return isnan (pt.x) || isnan (pt.y);
+}
+
+static bool
+isnan_3BD (BITCODE_3BD pt)
+{
+  return isnan (pt.x) || isnan (pt.y) || isnan (pt.z);
+}
+
 static void
 output_TEXT (Dwg_Object *obj)
 {
@@ -107,28 +125,36 @@ output_TEXT (Dwg_Object *obj)
   Dwg_Object_STYLE *style = o ? o->tio.object->tio.STYLE : NULL;
   BITCODE_2DPOINT pt;
 
+  if (!text->text_value)
+    return;
+  if (isnan_2BD (text->insertion_pt) || isnan_3BD (text->extrusion))
+    return;
   if (dwg->header.version >= R_2007)
     escaped = htmlwescape ((BITCODE_TU)text->text_value);
   else
     escaped = htmlescape (text->text_value, (int)dwg->header.codepage);
-  if (style && style->font_name && *style->font_name
+
+  if (style && o->fixedtype == DWG_TYPE_STYLE && style->font_name
+      && *style->font_name
 #ifdef HAVE_STRCASESTR
       && strcasestr (style->font_name, ".ttf")
 #else
-      && (strstr (style->font_name, ".ttf") || strstr (style->font_name, ".TTF"))
+      && (strstr (style->font_name, ".ttf")
+          || strstr (style->font_name, ".TTF"))
 #endif
-      )
+  )
     {
 #ifdef HAVE_STRCASESTR
       if (strcasestr (style->font_name, "Arial"))
 #else
-      if ((strstr (style->font_name, "arial")) || strstr (style->font_name, "Arial"))
+      if ((strstr (style->font_name, "arial"))
+          || strstr (style->font_name, "Arial"))
 #endif
         {
           fontfamily = "Arial";
         }
       else
-          fontfamily = "Verdana";
+        fontfamily = "Verdana";
     }
   else
     fontfamily = "Courier";
@@ -136,9 +162,8 @@ output_TEXT (Dwg_Object *obj)
   transform_OCS_2d (&pt, text->insertion_pt, text->extrusion);
   printf ("\t<text id=\"dwg-object-%d\" x=\"%f\" y=\"%f\" "
           "font-family=\"%s\" font-size=\"%f\" fill=\"blue\">%s</text>\n",
-          obj->index, transform_X (pt.x), transform_Y (pt.y),
-          fontfamily, text->height /* fontsize */,
-          escaped ? escaped : "");
+          obj->index, transform_X (pt.x), transform_Y (pt.y), fontfamily,
+          text->height /* fontsize */, escaped ? escaped : "");
   free (escaped);
 }
 
@@ -148,6 +173,9 @@ output_LINE (Dwg_Object *obj)
   Dwg_Entity_LINE *line = obj->tio.entity->tio.LINE;
   BITCODE_3DPOINT start, end;
 
+  if (isnan_3BD (line->start) || isnan_3BD (line->end)
+      || isnan_3BD (line->extrusion))
+    return;
   transform_OCS (&start, line->start, line->extrusion);
   transform_OCS (&end, line->end, line->extrusion);
   printf ("\t<path id=\"dwg-object-%d\" d=\"M %f,%f %f,%f\" "
@@ -162,6 +190,9 @@ output_CIRCLE (Dwg_Object *obj)
   Dwg_Entity_CIRCLE *circle = obj->tio.entity->tio.CIRCLE;
   BITCODE_3DPOINT center;
 
+  if (isnan_3BD (circle->center) || isnan_3BD (circle->extrusion)
+      || isnan (circle->radius))
+    return;
   transform_OCS (&center, circle->center, circle->extrusion);
   printf ("\t<circle id=\"dwg-object-%d\" cx=\"%f\" cy=\"%f\" r=\"%f\" "
           "fill=\"none\" stroke=\"blue\" stroke-width=\"0.1px\" />\n",
@@ -177,6 +208,10 @@ output_ARC (Dwg_Object *obj)
   double x_start, y_start, x_end, y_end;
   int large_arc;
 
+  if (isnan_3BD (arc->center) || isnan_3BD (arc->extrusion)
+      || isnan (arc->radius) || isnan (arc->start_angle)
+      || isnan (arc->start_angle))
+    return;
   transform_OCS (&center, arc->center, arc->extrusion);
 
   x_start = center.x + arc->radius * cos (arc->start_angle);
@@ -206,18 +241,20 @@ output_POLYLINE_2D (Dwg_Object *obj)
       dwg_point_2d *pts = dwg_object_polyline_2d_get_points (obj, &error);
       BITCODE_RL j;
 
-      if (error)
+      if (error || isnan_2pt (pts[0]) || isnan_3BD (pline->extrusion))
         return;
       ptin.x = pts[0].x;
       ptin.y = pts[0].y;
       transform_OCS_2d (&pt, ptin, pline->extrusion);
-      printf ("\t<path id=\"dwg-object-%d\" d=\"M %f,%f",
-              obj->index, transform_X (pt.x), transform_Y (pt.y));
+      printf ("\t<path id=\"dwg-object-%d\" d=\"M %f,%f", obj->index,
+              transform_X (pt.x), transform_Y (pt.y));
       // TODO curve_types, C for Bezier
       for (j = 1; j < numpts; j++)
         {
           ptin.x = pts[j].x;
           ptin.y = pts[j].y;
+          if (isnan_2BD (ptin))
+            continue;
           transform_OCS_2d (&pt, ptin, pline->extrusion);
           // TODO bulge -> arc, widths
           printf (" L %f,%f", transform_X (pt.x), transform_Y (pt.y));
@@ -242,18 +279,20 @@ output_LWPOLYLINE (Dwg_Object *obj)
       dwg_point_2d *pts = dwg_ent_lwpline_get_points (pline, &error);
       BITCODE_RL j;
 
-      if (error)
+      if (error || isnan_2pt (pts[0]) || isnan_3BD (pline->extrusion))
         return;
       ptin.x = pts[0].x;
       ptin.y = pts[0].y;
       transform_OCS_2d (&pt, ptin, pline->extrusion);
-      printf ("\t<path id=\"dwg-object-%d\" d=\"M %f,%f",
-              obj->index, transform_X (pt.x), transform_Y (pt.y));
+      printf ("\t<path id=\"dwg-object-%d\" d=\"M %f,%f", obj->index,
+              transform_X (pt.x), transform_Y (pt.y));
       // TODO curve_types, C for Bezier
       for (j = 1; j < numpts; j++)
         {
           ptin.x = pts[j].x;
           ptin.y = pts[j].y;
+          if (isnan_2BD (ptin))
+            continue;
           transform_OCS_2d (&pt, ptin, pline->extrusion);
           // TODO bulge -> arc, widths
           printf (" L %f,%f", transform_X (pt.x), transform_Y (pt.y));
@@ -272,6 +311,9 @@ output_INSERT (Dwg_Object *obj)
   if (insert->block_header && insert->block_header->handleref.value)
     {
       BITCODE_3DPOINT ins_pt;
+      if (isnan_3BD (insert->ins_pt) || isnan_3BD (insert->extrusion)
+          || isnan (insert->rotation) || isnan_3BD (insert->scale))
+        return;
       transform_OCS (&ins_pt, insert->ins_pt, insert->extrusion);
       printf ("\t<use id=\"dwg-object-%d\" transform=\"translate(%f %f) "
               "rotate(%f) scale(%f %f)\" xlink:href=\"#symbol-%lX\" />"
@@ -392,6 +434,10 @@ output_SVG (Dwg_Data *dwg)
   // double scale_x = dx / (dwg_page_x_max(dwg) - dwg_page_x_min(dwg));
   // double scale_y = dy / (dwg_page_y_max(dwg) - dwg_page_y_min(dwg));
   // scale = 25.4 / 72; // pt:mm
+  if (isnan (dx))
+    dx = 100.0;
+  if (isnan (dy))
+    dy = 100.0;
   page_width = dx;
   page_height = dy;
   // scale *= (scale_x > scale_y ? scale_x : scale_y);
@@ -408,7 +454,7 @@ output_SVG (Dwg_Data *dwg)
           page_width, page_height);
 
   if ((ref = dwg_model_space_ref (dwg)))
-      output_BLOCK_HEADER (ref);
+    output_BLOCK_HEADER (ref);
   if ((ref = dwg_paper_space_ref (dwg)))
     output_BLOCK_HEADER (ref);
   printf ("\t<defs>\n");
@@ -418,7 +464,6 @@ output_SVG (Dwg_Data *dwg)
         output_BLOCK_HEADER (ref);
     }
   printf ("\t</defs>\n");
-
   printf ("</svg>\n");
 }
 
