@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2018 Free Software Foundation, Inc.                        */
+/*  Copyright (C) 2018-2019 Free Software Foundation, Inc.                   */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -10,37 +10,42 @@
 /*  along with this program.  If not, see <http://www.gnu.org/licenses/>.    */
 /*****************************************************************************/
 
-/* TODO: format readers, outfile logic
- * dwgwrite.c: write a DWG file from various input formats.
+/* dwgwrite.c: write a DWG file from various input formats.
  * written by Reini Urban
+ * TODO: format readers, loglevel
  */
 
 #include "../src/config.h"
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 // strings.h or string.h
 #ifdef AX_STRCASECMP_HEADER
 #  include AX_STRCASECMP_HEADER
 #endif
+#include <sys/stat.h>
+#include <unistd.h>
 #include <getopt.h>
 #ifdef HAVE_VALGRIND_VALGRIND_H
 #  include <valgrind/valgrind.h>
 #endif
 
 #include "dwg.h"
+#include "common.h"
 #include "bits.h"
 #include "suffix.inc"
 #include "in_json.h"
 #include "in_dxf.h"
 
 static int opts = 1;
+int overwrite = 0;
 
 static int help (void);
 
 static int
 usage (void)
 {
-  printf ("\nUsage: dwgwrite [-v[0-9]] [--as rNNNN] [-I FMT] [-o DWGFILE] "
+  printf ("\nUsage: dwgwrite [-v[0-9]] [-y] [--as rNNNN] [-I FMT] [-o DWGFILE] "
           "INFILE\n");
   return 1;
 }
@@ -54,7 +59,7 @@ static int
 help (void)
 {
   printf ("\nUsage: dwgwrite [OPTION]... [-o DWGFILE] INFILE\n");
-  printf ("Writes a DWG file from various input formats. Only r2000 for now\n"
+  printf ("Writes a DWG file from various input formats. Only r2000 for now.\n"
           "\n");
 #ifdef HAVE_GETOPT_LONG
   printf ("  -v[0-9], --verbose [0-9]  verbosity\n");
@@ -67,6 +72,7 @@ help (void)
   printf (
       "           Planned input formats: JSON, GeoJSON, YAML, XML/OGR, GPX\n");
   printf ("  -o dwgfile, --file        \n");
+  printf ("  -y, --overwrite           overwrite existing files\n");
   printf ("           --help           display this help and exit\n");
   printf ("           --version        output version information and exit\n"
           "\n");
@@ -81,6 +87,7 @@ help (void)
   printf ("              Planned input formats: JSON, GeoJSON, YAML, XML/OGR, "
           "GPX\n");
   printf ("  -o dwgfile\n");
+  printf ("  -y          overwrite existing files\n");
   printf ("  -h          display this help and exit\n");
   printf ("  -i          output version information and exit\n"
           "\n");
@@ -103,13 +110,15 @@ main (int argc, char *argv[])
   const char *version = NULL;
   Dwg_Version_Type dwg_version = R_INVALID;
   int c;
+  int free_outfile = 0;
 #ifdef HAVE_GETOPT_LONG
   int option_index = 0;
   static struct option long_options[]
       = { { "verbose", 1, &opts, 1 }, // optional
           { "format", 1, 0, 'I' },    { "file", 1, 0, 'o' },
           { "as", 1, 0, 'a' },        { "help", 0, 0, 0 },
-          { "version", 0, 0, 0 },     { NULL, 0, NULL, 0 } };
+          { "overwrite", 0, 0, 'y' }, { "version", 0, 0, 0 },
+          { NULL, 0, NULL, 0 } };
 #endif
 
   if (argc < 2)
@@ -118,10 +127,10 @@ main (int argc, char *argv[])
   while
 #ifdef HAVE_GETOPT_LONG
       ((c
-        = getopt_long (argc, argv, ":a:v::I:o:h", long_options, &option_index))
+        = getopt_long (argc, argv, "ya:v::I:o:h", long_options, &option_index))
        != -1)
 #else
-      ((c = getopt (argc, argv, ":a:v::I:o:hi")) != -1)
+      ((c = getopt (argc, argv, "ya:v::I:o:hi")) != -1)
 #endif
     {
       if (c == -1)
@@ -166,6 +175,9 @@ main (int argc, char *argv[])
         case 'I':
           fmt = optarg;
           break;
+        case 'y':
+          overwrite = 1;
+          break;
         case 'o':
           outfile = optarg;
           break;
@@ -215,6 +227,7 @@ main (int argc, char *argv[])
       infile = argv[i];
       if (!fmt)
         {
+#ifndef DISABLE_DXF
           if (strstr (infile, ".json") || strstr (infile, ".JSON"))
             fmt = (char *)"json";
           else if (strstr (infile, ".dxfb") || strstr (infile, ".DXFB"))
@@ -222,6 +235,7 @@ main (int argc, char *argv[])
           else if (strstr (infile, ".dxf") || strstr (infile, ".DXF"))
             fmt = (char *)"dxf";
           else
+#endif
             fprintf (stderr, "Unknown input format for '%s'\n", infile);
         }
     }
@@ -230,24 +244,57 @@ main (int argc, char *argv[])
   memset (&dwg, 0, sizeof (Dwg_Data));
   dwg.opts = opts;
   if (infile)
-    dat.fh = fopen (infile, "r");
+    {
+      struct stat attrib;
+      if (stat (infile, &attrib)) // not exists
+        {
+          fprintf (stderr, "Missing input file '%s'\n", infile);
+          exit (1);
+        }
+      dat.fh = fopen (infile, "r");
+      if (!dat.fh)
+        {
+          fprintf (stderr, "Could not read file '%s'\n", infile);
+          exit (1);
+        }
+      dat.size = attrib.st_size;
+    }
   else
     dat.fh = stdin;
 
-  /*if ((fmt && !strcasecmp(fmt, "json")) ||
-        (infline && !strcasecmp(infile, ".json")))
-    error = dwg_read_json(&dat, &dwg);
+#ifndef DISABLE_DXF
+  /* if ((fmt && !strcasecmp (fmt, "json"))
+      || (infile && !strcasecmp (infile, ".json")))
+    {
+      if (opts > 1)
+        printf ("Reading JSON file %s\n", infile ? infile : "from stdin");
+      error = dwg_read_json (&dat, &dwg);
+    }
   else */
   if ((fmt && !strcasecmp (fmt, "dxfb"))
       || (infile && !strcasecmp (infile, ".dxfb")))
-    error = dwg_read_dxfb (&dat, &dwg);
+    {
+      if (opts > 1)
+        fprintf (stderr, "Reading Binary DXF file %s\n",
+                 infile ? infile : "from stdin");
+      error = dwg_read_dxfb (&dat, &dwg);
+    }
   else if ((fmt && !strcasecmp (fmt, "dxf"))
            || (infile && !strcasecmp (infile, ".dxf")))
-    error = dxf_read_file (infile, &dwg); // ascii or binary
+    {
+      if (opts > 1)
+        fprintf (stderr, "Reading DXF file %s\n",
+                 infile ? infile : "from stdin");
+      if (infile)
+        error = dxf_read_file (infile, &dwg); // ascii or binary
+      else
+        error = dwg_read_dxf (&dat, &dwg);
+    }
   else
+#endif
     {
       if (fmt)
-        fprintf (stderr, "Invalid input format '%s'\n", fmt);
+        fprintf (stderr, "Invalid or unsupported input format '%s'\n", fmt);
       else if (infile)
         fprintf (stderr, "Missing input format for '%s'\n", infile);
       else
@@ -256,8 +303,15 @@ main (int argc, char *argv[])
         fclose (dat.fh);
       exit (1);
     }
-  if (infile)
+  if (infile && dat.fh)
     fclose (dat.fh);
+  if (error >= DWG_ERR_CRITICAL)
+    {
+      fprintf (stderr, "ERROR 0x%x\n", error);
+      if (error && opts > 2)
+        dwg_errstrings (error);
+      exit (1);
+    }
 
   if (dwg.header.from_version != dwg.header.version)
     dwg.header.from_version = dwg.header.version;
@@ -270,15 +324,82 @@ main (int argc, char *argv[])
       dat.version = dwg.header.version = R_2000;
     }
   if (!outfile)
-    outfile = suffix (infile, "dwg");
-  error |= dwg_write_file (outfile, &dwg);
-  // forget about valgrind. really huge DWG's need endlessly here.
+    {
+      outfile = suffix (infile, "dwg");
+      free_outfile = 1;
+    }
+
+  if (opts > 1)
+    fprintf (stderr, "Writing DWG file %s\n", outfile);
+
+  {
+    struct stat attrib;
+    if (!stat (outfile, &attrib)) // exists
+      {
+        if (!overwrite)
+          {
+            fprintf (stderr, "File not overwritten: %s, use -y.\n", outfile);
+            error |= DWG_ERR_IOERROR;
+          }
+        else
+          {
+            if (S_ISREG (attrib.st_mode) &&   // refuse to remove a directory
+                (access (outfile, W_OK) == 0) // writable
+#ifndef _WIN32
+                // refuse to remove a symlink. even with overwrite. security
+                && !S_ISLNK (attrib.st_mode)
+#endif
+            )
+              {
+                unlink (outfile);
+                dwg.opts |= opts;
+                error = dwg_write_file (outfile, &dwg);
+              }
+            else
+              {
+                fprintf (stderr, "Not writable file or symlink: %s\n",
+                         outfile);
+                error |= DWG_ERR_IOERROR;
+              }
+          }
+      }
+    else
+      {
+        dwg.opts |= opts;
+        error = dwg_write_file (outfile, &dwg);
+      }
+  }
+
+  // forget about leaks. really huge DWG's need endlessly here.
   if ((dwg.header.version && dwg.num_objects < 1000)
+#if defined __SANITIZE_ADDRESS__ || __has_feature(address_sanitizer)
+      || 1
+#endif
 #ifdef HAVE_VALGRIND_VALGRIND_H
       || (RUNNING_ON_VALGRIND)
 #endif
   )
     dwg_free (&dwg);
 
+  if (error >= DWG_ERR_CRITICAL)
+    {
+      fprintf (stderr, "ERROR 0x%x\n", error);
+      if (error && opts > 2)
+        dwg_errstrings (error);
+    }
+  else
+    {
+      if (opts > 1)
+        {
+          fprintf (stderr, "SUCCESS 0x%x\n", error);
+          if (error && opts > 2)
+            dwg_errstrings (error);
+        }
+      else
+        fprintf (stderr, "SUCCESS\n");
+    }
+
+  if (free_outfile)
+    free (outfile);
   return error >= DWG_ERR_CRITICAL ? 1 : 0;
 }
