@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2009,2018-2019 Free Software Foundation, Inc.              */
+/*  Copyright (C) 2009,2018-2020 Free Software Foundation, Inc.              */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -58,7 +58,7 @@ bit_advance_position (Bit_Chain *dat, long advance)
       // but allow pointing to the very end.
       if (dat->byte != dat->size - 1 || dat->bit != 0)
         {
-          loglevel = dat->opts & 0xf;
+          loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
           LOG_ERROR ("%s buffer overflow at pos %lu, size %lu, advance by %ld",
                      __FUNCTION__, dat->byte, dat->size, advance)
         }
@@ -68,7 +68,7 @@ bit_advance_position (Bit_Chain *dat, long advance)
     }
   if ((long)dat->byte + (endpos / 8) < 0)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("buffer underflow at pos %lu, size %lu, advance by %ld",
                  dat->byte, dat->size, advance)
       dat->byte = 0;
@@ -96,7 +96,7 @@ bit_set_position (Bit_Chain *dat, unsigned long bitpos)
   dat->bit = bitpos & 7;
   if (dat->byte > dat->size || (dat->byte == dat->size && dat->bit))
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("%s buffer overflow at %lu, have %lu", __FUNCTION__,
                  dat->byte, dat->size)
     }
@@ -124,7 +124,7 @@ bit_read_B (Bit_Chain *dat)
 
   if (dat->byte >= dat->size)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte)
 #ifdef DWG_ABORT
       if (++errors > DWG_ABORT_LIMIT)
@@ -165,7 +165,7 @@ bit_read_BB (Bit_Chain *dat)
 
   if (dat->byte >= dat->size)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte)
 #ifdef DWG_ABORT
       if (++errors > DWG_ABORT_LIMIT)
@@ -254,7 +254,7 @@ bit_write_3B (Bit_Chain *dat, unsigned char value)
 {
   if (value > 7)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("Invalid bit_write_3B value %d > 7", value)
       return;
     }
@@ -307,7 +307,7 @@ bit_read_RC (Bit_Chain *dat)
 
   if (dat->byte >= dat->size)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte)
 #ifdef DWG_ABORT
       if (++errors > DWG_ABORT_LIMIT)
@@ -328,7 +328,7 @@ bit_read_RC (Bit_Chain *dat)
         }
       else
         {
-          loglevel = dat->opts & 0xf;
+          loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
           LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte + 1)
           return 0;
         }
@@ -567,7 +567,7 @@ bit_read_BL (Bit_Chain *dat)
     return 0;
   else /* if (two_bit_code == 3) */
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("bit_read_BL: unexpected 2-bit code: '11'")
       return 256;
     }
@@ -769,7 +769,7 @@ bit_read_BD (Bit_Chain *dat)
     return 0.0;
   else /* if (two_bit_code == 3) */
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("bit_read_BD: unexpected 2-bit code: '11'")
       return bit_nan ();
     }
@@ -847,7 +847,7 @@ bit_read_MC (Bit_Chain *dat)
       result |= ((BITCODE_UMC)byte[i]) << j;
     }
 
-  loglevel = dat->opts & 0xf;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
   LOG_ERROR (
       "bit_read_MC: error parsing modular char. i=%d, j=%d, result=0x%lx,\n"
       " @%lu.@%u: [0x%x 0x%x 0x%x 0x%x 0x%x]",
@@ -883,7 +883,7 @@ bit_write_MC (Bit_Chain *dat, BITCODE_MC val)
     if (byte[i] & 0x7f)
       break;
 
-  if (byte[i] & 0x40)
+  if (byte[i] & 0x40 && i > 0)
     i--;
   byte[i] &= 0x7f;
   if (negative)
@@ -919,7 +919,7 @@ bit_read_UMC (Bit_Chain *dat)
       result |= ((BITCODE_UMC)byte[i]) << j;
     }
 
-  loglevel = dat->opts & 0xf;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
   LOG_ERROR (
       "bit_read_UMC: error parsing modular char, i=%d,j=%d,result=0x%lx", i, j,
       result)
@@ -953,7 +953,7 @@ bit_write_UMC (Bit_Chain *dat, BITCODE_UMC val)
     if (byte[i] & 0x7f)
       break;
 
-  if (byte[i] & 0x40)
+  if (byte[i] & 0x40 && i > 0)
     i--;
   byte[i] &= 0x7f;
   for (j = 4; j >= i; j--)
@@ -982,7 +982,7 @@ bit_read_MS (Bit_Chain *dat)
         word[i] &= 0x7fff;
       result |= ((BITCODE_MS)word[i]) << j;
     }
-  loglevel = dat->opts & 0xf;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
   LOG_ERROR ("bit_read_MS: error parsing modular short, i=%d,j=%d", i, j)
   return 0; /* error... */
 }
@@ -1164,7 +1164,7 @@ bit_read_H (Bit_Chain *restrict dat, Dwg_Handle *restrict handle)
   handle->value = 0;
   if (handle->size > sizeof(BITCODE_RC *) || handle->code > 14)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_WARN ("Invalid handle-reference, longer than 8 bytes: " FORMAT_H,
                 ARGS_H (*handle))
       return DWG_ERR_INVALIDHANDLE;
@@ -1239,17 +1239,15 @@ uint16_t
 bit_read_CRC (Bit_Chain *dat)
 {
   uint16_t result;
-  long unsigned int start_address;
-  loglevel = dat->opts & 0xf;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
 
   if (dat->bit > 0)
     {
       dat->byte++;
       dat->bit = 0;
     }
-  start_address = dat->byte;
   result = bit_read_RS (dat);
-  LOG_TRACE ("read CRC at %lu: %04X\n", start_address, result)
+  LOG_TRACE ("read CRC at %lu: %04X\n", dat->byte, result)
 
   return result;
 }
@@ -1262,7 +1260,7 @@ bit_check_CRC (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
   uint16_t calculated;
   uint16_t read;
   long size;
-  loglevel = dat->opts & 0xf;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
 
   if (dat->bit > 0)
     {
@@ -1272,12 +1270,11 @@ bit_check_CRC (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
 
   if (start_address > dat->byte || dat->byte >= dat->size)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("%s buffer overflow at pos %lu-%lu, size %lu", __FUNCTION__,
                  start_address, dat->byte, dat->size)
       return 0;
     }
-
   size = dat->byte - start_address;
   calculated = bit_calc_CRC (seed, &dat->chain[start_address], size);
   read = bit_read_RS (dat);
@@ -1303,11 +1300,18 @@ bit_write_CRC (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
 {
   uint16_t crc;
   long size;
-  loglevel = dat->opts & 0xf;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
 
   while (dat->bit > 0)
     bit_write_B (dat, 0);
 
+  if (start_address > dat->byte || dat->byte >= dat->size)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at pos %lu-%lu, size %lu", __FUNCTION__,
+                 start_address, dat->byte, dat->size)
+      return 0;
+    }
   size = dat->byte - start_address;
   crc = bit_calc_CRC (seed, &dat->chain[start_address], size);
 
@@ -1322,14 +1326,21 @@ bit_write_CRC_LE (Bit_Chain *dat, long unsigned int start_address, uint16_t seed
 {
   uint16_t crc;
   long size;
-  loglevel = dat->opts & 0xf;
 
   while (dat->bit > 0)
     bit_write_B (dat, 0);
 
+  if (start_address > dat->byte || dat->byte >= dat->size)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at pos %lu-%lu, size %lu", __FUNCTION__,
+                 start_address, dat->byte, dat->size)
+      return 0;
+    }
   size = dat->byte - start_address;
   crc = bit_calc_CRC (seed, &dat->chain[start_address], size);
 
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
   LOG_TRACE ("write CRC %04X from %lu-%lu = %ld\n", crc, start_address,
              dat->byte, size);
   bit_write_RS_LE (dat, crc);
@@ -1337,11 +1348,27 @@ bit_write_CRC_LE (Bit_Chain *dat, long unsigned int start_address, uint16_t seed
 }
 
 void
-bit_read_fixed (Bit_Chain *restrict dat, BITCODE_RC *restrict dest, int length)
+bit_read_fixed (Bit_Chain *restrict dat, BITCODE_RC *restrict dest,
+                unsigned int length)
 {
-  for (int i = 0; i < length; i++)
+  if (dat->byte + length >= dat->size)
     {
-      dest[i] = bit_read_RC (dat);
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at pos %lu, size %lu", __FUNCTION__,
+                 dat->byte, dat->size)
+      return;
+    }
+  if (dat->bit == 0)
+    {
+      memcpy (dest, &dat->chain[dat->byte], length);
+      dat->byte += length;
+    }
+  else
+    {
+      for (unsigned int i = 0; i < length; i++)
+        {
+          dest[i] = bit_read_RC (dat);
+        }
     }
 }
 
@@ -1349,10 +1376,11 @@ bit_read_fixed (Bit_Chain *restrict dat, BITCODE_RC *restrict dest, int length)
  *  After usage, the allocated memory must be properly freed.
  *  preR11
  */
+ATTRIBUTE_MALLOC
 BITCODE_TF
-bit_read_TF (Bit_Chain *restrict dat, int length)
+bit_read_TF (Bit_Chain *restrict dat, unsigned int length)
 {
-  BITCODE_RC *chain = malloc (length + 1);
+  BITCODE_RC *chain = calloc (length + 1, 1);
 
   bit_read_fixed (dat, chain, length);
   chain[length] = '\0';
@@ -1363,11 +1391,18 @@ bit_read_TF (Bit_Chain *restrict dat, int length)
 /** Write fixed text. Without ending \0
  */
 void
-bit_write_TF (Bit_Chain *restrict dat, BITCODE_TF restrict chain, int length)
+bit_write_TF (Bit_Chain *restrict dat, BITCODE_TF restrict chain, unsigned int length)
 {
-  int i;
-  for (i = 0; i < length; i++)
-    bit_write_RC (dat, (BITCODE_RC)chain[i]);
+  if (dat->bit == 0 && dat->byte + length < dat->size)
+    {
+      memcpy (&dat->chain[dat->byte], chain, length);
+      dat->byte += length;
+    }
+  else
+    {
+      for (unsigned int i = 0; i < length; i++)
+        bit_write_RC (dat, (BITCODE_RC)chain[i]);
+    }
 }
 
 /** Read simple text. After usage, the allocated memory must be properly freed.
@@ -1382,7 +1417,7 @@ bit_read_TV (Bit_Chain *restrict dat)
   length = bit_read_BS (dat);
   if (dat->byte + length >= dat->size)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("%s buffer overflow at %lu, length %u", __FUNCTION__,
                  dat->byte, length)
       return NULL;
@@ -1429,7 +1464,7 @@ bit_read_TU (Bit_Chain *restrict dat)
   length = bit_read_BS (dat);
   if (dat->byte + (length * 2) >= dat->size)
     {
-      loglevel = dat->opts & 0xf;
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("%s buffer overflow at %lu, length %u", __FUNCTION__,
                  dat->byte, length)
       return NULL;
@@ -1594,7 +1629,7 @@ bit_convert_TU (BITCODE_TU restrict wstr)
           str[i++] = (c & 0x3F) | 0x80;
         }
       else
-        fprintf (stderr, "ERROR: overlarge unicode codepoint U+%0X", c);
+        HANDLER (OUTPUT, "ERROR: overlarge unicode codepoint U+%0X", c);
      */
     }
   str[i] = '\0';
@@ -1611,8 +1646,9 @@ bit_utf8_to_TU (char *restrict str)
   unsigned char c;
 
   wstr = malloc (2 * (len + 1));
-  while ((c = *str++))
+  while (len >= 0 && (c = *str++))
     {
+      len--;
       if (c < 128)
         {
           wstr[i++] = c;
@@ -1620,25 +1656,31 @@ bit_utf8_to_TU (char *restrict str)
       else if ((c & 0xe0) == 0xc0)
         {
           /* ignore invalid utf8 for now */
-          wstr[i++] = ((c & 0x1f) << 6) | (str[1] & 0x3f);
+          if (len >= 1)
+            wstr[i++] = ((c & 0x1f) << 6) | (str[1] & 0x3f);
+          len--;
           str++;
         }
       else if ((c & 0xf0) == 0xe0)
         {
           /* ignore invalid utf8? */
-          if ((unsigned char)str[1] < 0x80 || (unsigned char)str[1] > 0xBF
-              || (unsigned char)str[2] < 0x80 || (unsigned char)str[2] > 0xBF)
+          if (len >= 2 &&
+              ((unsigned char)str[1] < 0x80 || (unsigned char)str[1] > 0xBF
+               || (unsigned char)str[2] < 0x80 || (unsigned char)str[2] > 0xBF))
             {
               LOG_WARN ("utf-8: BAD_CONTINUATION_BYTE %s", str);
             }
-          if (c == 0xe0 && (unsigned char)str[1] < 0xa0)
+          if (len >= 1 && c == 0xe0 && (unsigned char)str[1] < 0xa0)
             {
               LOG_WARN ("utf-8: NON_SHORTEST %s", str);
             }
-          wstr[i++]
+          if (len >= 2)
+            wstr[i++]
               = ((c & 0x0f) << 12) | ((str[1] & 0x3f) << 6) | (str[2] & 0x3f);
           str++;
           str++;
+          len--;
+          len--;
         }
       /* everything above 0xf0 exceeds ucs-2, 4-6 byte seqs */
     }
@@ -1837,7 +1879,9 @@ bit_search_sentinel (Bit_Chain *dat, unsigned char sentinel[16])
 {
   long unsigned int i, j;
 
-  for (i = 0; i < dat->size; i++)
+  if (dat->size < 16) // too short
+    return 0;
+  for (i = 0; i < dat->size - 16; i++)
     {
       for (j = 0; j < 16; j++)
         {
@@ -1882,6 +1926,7 @@ bit_chain_alloc (Bit_Chain *dat)
     {
       dat->chain
           = (unsigned char *)realloc (dat->chain, dat->size + CHAIN_BLOCK);
+      memset (&dat->chain[dat->size], 0, CHAIN_BLOCK);
       dat->size += CHAIN_BLOCK;
     }
 }
@@ -1925,7 +1970,7 @@ bit_write_bits (Bit_Chain *restrict dat, const char *restrict bits)
         bit_write_B (dat, *p != '0');
       else
         {
-          fprintf (stderr, "Invalid binary input %s\n", p);
+          HANDLER (OUTPUT, "ERROR: Invalid binary input %s\n", p);
           return;
         }
     }
@@ -1967,7 +2012,7 @@ bit_write_hexbits (Bit_Chain *restrict dat, const char *restrict bytes)
             }
           else
             {
-              fprintf (stderr, "Invalid hex input %s\n", p);
+              HANDLER (OUTPUT, "ERROR: Invalid hex input %s\n", p);
               return 0;
             }
         }
@@ -1996,8 +2041,8 @@ bit_fprint_bits (FILE *fp, unsigned char *bits, long unsigned int size)
       unsigned char bit = i % 8;
       unsigned char result = (bits[i / 8] & (0x80 >> bit)) >> (7 - bit);
       if (i && !bit)
-        printf (" ");
-      fprintf (fp, "%d", result ? 1 : 0);
+        HANDLER (fp, " ");
+      HANDLER (fp, "%d", result ? 1 : 0);
       // fprintf(fp, "%d", BIT(bits, i) ? 1 : 0);
     }
 }
