@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2018-2019 Free Software Foundation, Inc.                   */
+/*  Copyright (C) 2018-2020 Free Software Foundation, Inc.                   */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -23,6 +23,7 @@
 
 #include "common.h"
 #include "bits.h"
+//#include "myalloca.h"
 #include "dwg.h"
 #include "decode.h"
 #include "out_dxf.h"
@@ -43,8 +44,8 @@ static int dwg_dxfb_object (Bit_Chain *restrict dat,
 static int dxfb_3dsolid (Bit_Chain *restrict dat,
                          const Dwg_Object *restrict obj,
                          Dwg_Entity_3DSOLID *restrict _obj);
-static int dxfb_block_write (Bit_Chain *restrict dat, Dwg_Object *restrict hdr,
-                             int *restrict i);
+static int dxfb_block_write (Bit_Chain *restrict dat, const Dwg_Object *restrict mspace,
+                             Dwg_Object *restrict hdr, int *restrict i);
 static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
                                   const Dwg_Object *restrict obj,
                                   char *restrict name, const int dxf);
@@ -104,7 +105,7 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
 #define VALUE_TFF(str, dxf) VALUE_TV (str, dxf)
 #define VALUE_BINARY(value, size, dxf)                                        \
   {                                                                           \
-    long len = size;                                                          \
+    long len = (long)size;                                                    \
     do                                                                        \
       {                                                                       \
         short j;                                                              \
@@ -222,11 +223,11 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
     }
 
 /*
-#define VALUE(code, value)                   \
-  {\
-    GCC_DIAG_IGNORE(-Wformat-nonliteral) \
-    snprintf (buf, 4096, "%3i\n%s\n", code, dxfb_format (code));\
-    fprintf(dat->fh, buf, value);\
+#define VALUE(code, value) \
+  { \
+    GCC30_DIAG_IGNORE (-Wformat-nonliteral) \
+    snprintf (buf, 4096, "%3i\n%s\n", code, dxfb_format (code)); \
+    fprintf(dat->fh, buf, value); \
     GCC_DIAG_RESTORE \
   }
  */
@@ -301,8 +302,10 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
 
 #define HEADER_B(nam, dxf) HEADER_RC (nam, dxf)
 #define HEADER_BS(nam, dxf) HEADER_RS (nam, dxf)
+#define HEADER_BSd(nam, dxf) HEADER_RS (nam, dxf)
 #define HEADER_BD(nam, dxf) HEADER_RD (nam, dxf)
 #define HEADER_BL(nam, dxf) HEADER_RL (nam, dxf)
+#define HEADER_BLd(nam, dxf) HEADER_RL (nam, dxf)
 
 #define VALUE_H(value, dxf)                                                   \
   {                                                                           \
@@ -341,14 +344,18 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
   HEADER_9 (nam);                                                             \
   VALUE_HANDLE_NAME (dwg->header_vars.nam, dxf, table)
 
-#define HEADER_RLL(nam, dxf)                                                  \
+#define FIELD_BLL(nam, dxf)                                                   \
   {                                                                           \
-    BITCODE_RLL s = _obj->nam;                                                \
-    GROUP (9);                                                                \
-    fprintf (dat->fh, "$%s%c", #nam, 0);                                      \
+    BITCODE_BLL s = _obj->nam;                                                \
     GROUP (dxf);                                                              \
     fwrite (&s, sizeof (BITCODE_RLL), 1, dat->fh);                            \
   }
+#define HEADER_RLL(nam, dxf)                                                  \
+  {                                                                           \
+    GROUP (9);                                                                \
+    fprintf (dat->fh, "$%s%c", #nam, 0);                                      \
+    FIELD_BLL (nam, dxf);                                                     \
+  }
 
 #define FIELD_MC(nam, dxf) FIELD_RC (nam, dxf)
 #define FIELD_MS(nam, dxf) FIELD_RS (nam, dxf)
@@ -432,13 +439,13 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
 // reads data of the type indicated by 'type' 'size' times and stores
 // it all in the vector called 'nam'.
 #define FIELD_VECTOR_N(nam, type, size, dxf)                                  \
-  if (dxf)                                                                    \
+  if (dxf && _obj->nam)                                                       \
     {                                                                         \
       for (vcount = 0; vcount < (BITCODE_BL)size; vcount++)                   \
         VALUE (_obj->nam[vcount], type, dxf);                                 \
     }
 #define FIELD_VECTOR_T(nam, size, dxf)                                        \
-  if (dxf)                                                                    \
+  if (dxf && _obj->nam)                                                       \
     {                                                                         \
       PRE (R_2007)                                                            \
       {                                                                       \
@@ -456,7 +463,7 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
   FIELD_VECTOR_N (nam, type, _obj->size, dxf)
 
 #define FIELD_2RD_VECTOR(nam, size, dxf)                                      \
-  if (dxf)                                                                    \
+  if (dxf && _obj->nam)                                                       \
     {                                                                         \
       for (vcount = 0; vcount < (BITCODE_BL)_obj->size; vcount++)             \
         {                                                                     \
@@ -465,7 +472,7 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
     }
 
 #define FIELD_2DD_VECTOR(nam, size, dxf)                                      \
-  if (dxf)                                                                    \
+  if (dxf && _obj->nam)                                                       \
     {                                                                         \
       FIELD_2RD (nam[0], 0);                                                  \
       for (vcount = 1; vcount < (BITCODE_BL)_obj->size; vcount++)             \
@@ -476,7 +483,7 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
     }
 
 #define FIELD_3DPOINT_VECTOR(nam, size, dxf)                                  \
-  if (dxf)                                                                    \
+  if (dxf && _obj->nam)                                                       \
     {                                                                         \
       for (vcount = 0; vcount < (BITCODE_BL)_obj->size; vcount++)             \
         {                                                                     \
@@ -485,7 +492,7 @@ static void dxfb_cvt_tablerecord (Bit_Chain *restrict dat,
     }
 
 #define HANDLE_VECTOR_N(nam, size, code, dxf)                                 \
-  if (dxf)                                                                    \
+  if (dxf && _obj->nam)                                                       \
     {                                                                         \
       for (vcount = 0; vcount < (BITCODE_BL)size; vcount++)                   \
         {                                                                     \
@@ -571,21 +578,21 @@ static int dwg_dxfb_TABLECONTENT (Bit_Chain *restrict dat,
                                const Dwg_Object *restrict obj)                \
   {                                                                           \
     BITCODE_BL vcount, rcount1, rcount2, rcount3, rcount4;                    \
+    int error = 0;                                                            \
     Dwg_Data *dwg = obj->parent;                                              \
     Bit_Chain *str_dat = dat;                                                 \
     Dwg_Entity_##token *ent, *_obj;                                           \
     Dwg_Object_Entity *_ent;                                                  \
-    int error = 0;                                                            \
     if (obj->fixedtype != DWG_TYPE_##token)                                   \
       {                                                                       \
         LOG_ERROR ("Invalid type 0x%x, expected 0x%x %s", obj->fixedtype,     \
                    DWG_TYPE_##token, #token);                                 \
         return DWG_ERR_INVALIDTYPE;                                           \
       }                                                                       \
-    if (!strcmp (#token, "GEOPOSITIONMARKER"))                                \
+    if (strEQc (#token, "GEOPOSITIONMARKER"))                                 \
       RECORD (POSITIONMARKER)                                                 \
     else if (dat->version < R_13 && strlen (#token) == 10                     \
-             && !strcmp (#token, "LWPOLYLINE"))                               \
+             && strEQc (#token, "LWPOLYLINE"))                                \
       RECORD (POLYLINE)                                                       \
     else if (strlen (#token) > 11 && !memcmp (#token, "DIMENSION_", 11))      \
       RECORD (DIMENSION)                                                      \
@@ -593,7 +600,7 @@ static int dwg_dxfb_TABLECONTENT (Bit_Chain *restrict dat,
       RECORD (POLYLINE)                                                       \
     else if (strlen (#token) > 7 && !memcmp (#token, "VERTEX_", 7))           \
       RECORD (VERTEX)                                                         \
-    else if (dat->version >= R_2010 && !strcmp (#token, "TABLE"))             \
+    else if (dat->version >= R_2010 && strEQc (#token, "TABLE"))              \
       {                                                                       \
         RECORD (ACAD_TABLE);                                                  \
         return dwg_dxfb_TABLECONTENT (dat, obj);                              \
@@ -610,8 +617,7 @@ static int dwg_dxfb_TABLECONTENT (Bit_Chain *restrict dat,
     SINCE (R_11)                                                              \
     {                                                                         \
       uint32_t i = (uint32_t)obj->handle.value;                               \
-      LOG_TRACE ("Entity handle: %d.%d.%lX\n", obj->handle.code,              \
-                 obj->handle.size, obj->handle.value)                         \
+      LOG_TRACE ("Entity handle: " FORMAT_H "\n", ARGS_H (obj->handle))       \
       GROUP (330);                                                            \
       fwrite (&i, sizeof (uint32_t), 1, dat->fh);                             \
     }                                                                         \
@@ -623,7 +629,7 @@ static int dwg_dxfb_TABLECONTENT (Bit_Chain *restrict dat,
     }
 
 #define DWG_ENTITY_END                                                        \
-  return 0;                                                                   \
+    return error;                                                             \
   }
 
 #define DWG_OBJECT(token)                                                     \
@@ -631,10 +637,10 @@ static int dwg_dxfb_TABLECONTENT (Bit_Chain *restrict dat,
                                const Dwg_Object *restrict obj)                \
   {                                                                           \
     BITCODE_BL vcount, rcount1, rcount2, rcount3, rcount4;                    \
+    int error = 0;                                                            \
     Bit_Chain *hdl_dat = dat;                                                 \
     Dwg_Data *dwg = obj->parent;                                              \
     Dwg_Object_##token *_obj;                                                 \
-    int error = 0;                                                            \
     LOG_INFO ("Object " #token ":\n")                                         \
     if (obj->fixedtype != DWG_TYPE_##token)                                   \
       {                                                                       \
@@ -662,10 +668,15 @@ static int dwg_dxfb_TABLECONTENT (Bit_Chain *restrict dat,
           _XDICOBJHANDLE (3);                                                 \
           _REACTORS (4);                                                      \
         }                                                                     \
-      }
+        SINCE (R_14)                                                          \
+        {                                                                     \
+          VALUE_HANDLE (obj->tio.object->ownerhandle, ownerhandle, 3, 330);   \
+        }                                                                     \
+      }                                                                       \
+    LOG_TRACE ("Object handle: " FORMAT_H "\n", ARGS_H (obj->handle))
 
 #define DWG_OBJECT_END                                                        \
-  return 0;                                                                   \
+    return error;                                                             \
   }
 
 static void
@@ -746,26 +757,26 @@ dxfb_cvt_tablerecord (Bit_Chain *restrict dat, const Dwg_Object *restrict obj,
         }
       if (dat->from_version >= R_2000 && dat->version < R_2000)
         { // convert the other way round, from newer to older
-          if (!strcmp (name, "Standard"))
+          if (strEQc (name, "Standard"))
             VALUE_TV ("STANDARD", dxf)
-          else if (!strcmp (name, "ByLayer"))
+          else if (strEQc (name, "ByLayer"))
             VALUE_TV ("BYLAYER", dxf)
-          else if (!strcmp (name, "ByBlock"))
+          else if (strEQc (name, "ByBlock"))
             VALUE_TV ("BYBLOCK", dxf)
-          else if (!strcmp (name, "*Active"))
+          else if (strEQc (name, "*Active"))
             VALUE_TV ("*ACTIVE", dxf)
           else
             VALUE_TV (name, dxf)
         }
       else
         { // convert some standard names
-          if (dat->version >= R_2000 && !strcmp (name, "STANDARD"))
+          if (dat->version >= R_2000 && strEQc (name, "STANDARD"))
             VALUE_TV ("Standard", dxf)
-          else if (dat->version >= R_2000 && !strcmp (name, "BYLAYER"))
+          else if (dat->version >= R_2000 && strEQc (name, "BYLAYER"))
             VALUE_TV ("ByLayer", dxf)
-          else if (dat->version >= R_2000 && !strcmp (name, "BYBLOCK"))
+          else if (dat->version >= R_2000 && strEQc (name, "BYBLOCK"))
             VALUE_TV ("ByBlock", dxf)
-          else if (dat->version >= R_2000 && !strcmp (name, "*ACTIVE"))
+          else if (dat->version >= R_2000 && strEQc (name, "*ACTIVE"))
             VALUE_TV ("*Active", dxf)
           else
             VALUE_TV (name, dxf)
@@ -804,9 +815,9 @@ dxfb_cvt_blockname (Bit_Chain *restrict dat, char *restrict name,
     {
       if (strlen (name) < 10)
         VALUE_TV (name, dxf)
-      else if (!strcmp (name, "*Model_Space"))
+      else if (strEQc (name, "*Model_Space"))
         VALUE_TV ("$MODEL_SPACE", dxf)
-      else if (!strcmp (name, "*Paper_Space"))
+      else if (strEQc (name, "*Paper_Space"))
         VALUE_TV ("$PAPER_SPACE", dxf)
       else if (!memcmp (name, "*Paper_Space", sizeof ("*Paper_Space") - 1))
         {
@@ -820,9 +831,9 @@ dxfb_cvt_blockname (Bit_Chain *restrict dat, char *restrict name,
     {
       if (strlen (name) < 10)
         VALUE_TV (name, dxf)
-      else if (!strcmp (name, "$MODEL_SPACE"))
+      else if (strEQc (name, "$MODEL_SPACE"))
         VALUE_TV ("*Model_Space", dxf)
-      else if (!strcmp (name, "$PAPER_SPACE"))
+      else if (strEQc (name, "$PAPER_SPACE"))
         VALUE_TV ("*Paper_Space", dxf)
       else if (!memcmp (name, "$PAPER_SPACE", sizeof ("$PAPER_SPACE") - 1))
         {
@@ -836,7 +847,9 @@ dxfb_cvt_blockname (Bit_Chain *restrict dat, char *restrict name,
     free (name);
 }
 
-// 5 written here first
+#define START_OBJECT_HANDLE_STREAM
+
+// Handle 5 written here first
 #define COMMON_TABLE_CONTROL_FLAGS                                            \
   SINCE (R_13)                                                                \
   {                                                                           \
@@ -845,20 +858,23 @@ dxfb_cvt_blockname (Bit_Chain *restrict dat, char *restrict name,
     fwrite (&_i, sizeof (uint32_t), 1, dat->fh);                              \
   }                                                                           \
   SINCE (R_14)                                                                \
-  VALUE_H (_ctrl->null_handle, 330);                                          \
-  SINCE (R_13)                                                                \
-  VALUE_TV ("AcDbSymbolTable", 100)
+  {                                                                           \
+    VALUE_HANDLE (ctrl->tio.object->ownerhandle, ownerhandle, 3, 330);        \
+  }                                                                           \
+  SINCE (R_13) { VALUE_TV ("AcDbSymbolTable", 100); }
 
-#define COMMON_TABLE_FLAGS(owner, acdbname)                                   \
-  SINCE (R_14)                                                                \
+#define COMMON_TABLE_FLAGS(acdbname)                                          \
   /* TODO: ACAD_XDICTIONARY */                                                \
-  FIELD_HANDLE (owner, 4, 330);                                               \
+  SINCE (R_14)                                                                \
+  {                                                                           \
+    VALUE_HANDLE (obj->tio.object->ownerhandle, ownerhandle, 3, 330);         \
+  }                                                                           \
   SINCE (R_13)                                                                \
   {                                                                           \
     VALUE_TV ("AcDbSymbolTableRecord", 100)                                   \
     VALUE_TV ("AcDb" #acdbname "TableRecord", 100)                            \
   }                                                                           \
-  if (!strcmp (#acdbname, "Block") && dat->version >= R_13)                   \
+  if (strEQc (#acdbname, "Block") && dat->version >= R_13)                    \
     {                                                                         \
       Dwg_Object *blk = dwg_ref_object (                                      \
           dwg, ((Dwg_Object_BLOCK_HEADER *)_obj)->block_entity);              \
@@ -886,9 +902,11 @@ dxfb_cvt_blockname (Bit_Chain *restrict dat, char *restrict name,
     VALUE_TV ("*", 2)                                                         \
   FIELD_RC (flag, 70)
 
-#define LAYER_TABLE_FLAGS(owner, acdbname)                                    \
+#define LAYER_TABLE_FLAGS(acdbname)                                           \
   SINCE (R_14)                                                                \
-  FIELD_HANDLE (owner, 4, 330);                                               \
+  {                                                                           \
+    VALUE_HANDLE (obj->tio.object->ownerhandle, ownerhandle, 3, 330);         \
+  }                                                                           \
   SINCE (R_13)                                                                \
   {                                                                           \
     VALUE_TV ("AcDbSymbolTableRecord", 100)                                   \
@@ -904,7 +922,7 @@ static int
 dxfb_3dsolid (Bit_Chain *restrict dat, const Dwg_Object *restrict obj,
               Dwg_Entity_3DSOLID *restrict _obj)
 {
-  Dwg_Data *dwg = obj->parent;
+  //Dwg_Data *dwg = obj->parent;
   unsigned long j;
   BITCODE_BL vcount, rcount1, rcount2;
   BITCODE_BL i;
@@ -960,9 +978,8 @@ dxfb_3dsolid (Bit_Chain *restrict dat, const Dwg_Object *restrict obj,
         }
       else // if (FIELD_VALUE(version)==2)
         {
-          // TODO
           LOG_ERROR ("TODO: Implement parsing of SAT file (version 2) "
-                     "in entities 37,38 and 39.\n");
+                     "in entities 37,38 and 39.");
         }
       /*
             FIELD_B (wireframe_data_present, 0);
@@ -1008,7 +1025,7 @@ dxfb_3dsolid (Bit_Chain *restrict dat, const Dwg_Object *restrict obj,
       if (!FIELD_VALUE (acis_empty_bit))
         {
           LOG_ERROR ("TODO: Implement parsing of ACIS data at the end "
-                     "of 3dsolid object parsing (acis_empty_bit==0).\n");
+                     "of 3dsolid object parsing (acis_empty_bit==0).");
         }
 
       SINCE (R_2007)
@@ -1040,7 +1057,9 @@ dwg_dxfb_variable_type (const Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
   // almost always false
   is_entity = dwg_class_is_entity (klass);
 
-#include "classes.inc"
+  // clang-format off
+  #include "classes.inc"
+  // clang-format on
 
   return DWG_ERR_UNHANDLEDCLASS;
 }
@@ -1069,7 +1088,7 @@ dwg_dxfb_variable_type (const Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
           o = dwg_next_object (o);                                            \
           if (!o)                                                             \
             return DWG_ERR_INVALIDHANDLE;                                     \
-          if (!strcmp (#token, "PFACE")                                       \
+          if (strEQc (#token, "PFACE")                                        \
               && o->fixedtype == DWG_TYPE_VERTEX_PFACE_FACE)                  \
             {                                                                 \
               error |= dwg_dxfb_VERTEX_PFACE_FACE (dat, o);                   \
@@ -1092,7 +1111,7 @@ dwg_dxfb_variable_type (const Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
       for (BITCODE_BL j = 0; j < _obj->num_owned; j++)                        \
         {                                                                     \
           o = _obj->vertex[j] ? _obj->vertex[j]->obj : NULL;                  \
-          if (!strcmp (#token, "PFACE") && o                                  \
+          if (strEQc (#token, "PFACE") && o                                   \
               && o->fixedtype == DWG_TYPE_VERTEX_PFACE_FACE)                  \
             {                                                                 \
               error |= dwg_dxfb_VERTEX_PFACE_FACE (dat, o);                   \
@@ -1109,8 +1128,13 @@ dwg_dxfb_variable_type (const Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
     }                                                                         \
     return error;                                                             \
   }
-decl_dxfb_process_VERTEX (2D) decl_dxfb_process_VERTEX (3D)
-    decl_dxfb_process_VERTEX (MESH) decl_dxfb_process_VERTEX (PFACE)
+
+// clang-format off
+decl_dxfb_process_VERTEX (2D)
+decl_dxfb_process_VERTEX (3D)
+decl_dxfb_process_VERTEX (MESH)
+decl_dxfb_process_VERTEX (PFACE)
+// clang-format on
 
 /* process seqend before attribs */
 #define decl_dxfb_process_INSERT(token)                                       \
@@ -1161,20 +1185,22 @@ decl_dxfb_process_VERTEX (2D) decl_dxfb_process_VERTEX (3D)
     }                                                                         \
     return error;                                                             \
   }
-        decl_dxfb_process_INSERT (INSERT) decl_dxfb_process_INSERT (MINSERT)
 
-            static int dwg_dxfb_object (Bit_Chain *restrict dat,
-                                        const Dwg_Object *restrict obj,
-                                        int *restrict i)
+// clang-format off
+decl_dxfb_process_INSERT (INSERT)
+decl_dxfb_process_INSERT (MINSERT)
+// clang-format on
+
+static int dwg_dxfb_object (Bit_Chain *restrict dat,
+                            const Dwg_Object *restrict obj,
+                            int *restrict i)
 {
   int error = 0;
   int minimal;
 
   if (!obj || !obj->parent)
     return DWG_ERR_INTERNALERROR;
-  if (obj->supertype == DWG_SUPERTYPE_UNKNOWN)
-    return 0;
-  minimal = obj->parent->opts & 0x10;
+  minimal = obj->parent->opts & DWG_OPTS_MINIMAL;
 
   switch (obj->type)
     {
@@ -1207,7 +1233,6 @@ decl_dxfb_process_VERTEX (2D) decl_dxfb_process_VERTEX (3D)
         else
           return error;
       }
-
     case DWG_TYPE_POLYLINE_2D:
       error = dwg_dxfb_POLYLINE_2D (dat, obj);
       return error | dxfb_process_VERTEX_2D (dat, obj, i);
@@ -1360,7 +1385,7 @@ decl_dxfb_process_VERTEX (2D) decl_dxfb_process_VERTEX (3D)
     case DWG_TYPE_LAYOUT:
       return minimal ? 0 : dwg_dxfb_LAYOUT (dat, obj);
     default:
-      if (obj->type == obj->parent->layout_number)
+      if (obj->type == obj->parent->layout_type)
         {
           return minimal ? 0 : dwg_dxfb_LAYOUT (dat, obj);
         }
@@ -1370,7 +1395,6 @@ decl_dxfb_process_VERTEX (2D) decl_dxfb_process_VERTEX (3D)
                                                   (Dwg_Object *)obj)))
         {
           Dwg_Data *dwg = obj->parent;
-          int is_entity;
           int j = obj->type - 500;
           Dwg_Class *klass = NULL;
 
@@ -1399,20 +1423,24 @@ dxfb_common_entity_handle_data (Bit_Chain *restrict dat,
   ent = obj->tio.entity;
   _obj = ent;
 
-#include "common_entity_handle_data.spec"
+  // clang-format off
+  #include "common_entity_handle_data.spec"
+  #include "common_entity_data.spec"
+  // clang-format on
 
   return error;
 }
 
 // see
 // https://www.autodesk.com/techpubs/autocad/acad2000/dxf/header_section_group_codes_dxf_02.htm
+AFL_GCC_TOOBIG
 static int
 dxfb_header_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
   Dwg_Header_Variables *_obj = &dwg->header_vars;
   Dwg_Object *obj = NULL;
   double ms;
-  const int minimal = dwg->opts & 0x10;
+  const int minimal = dwg->opts & DWG_OPTS_MINIMAL;
   const char *codepage = dxf_codepage (dwg->header.codepage, dwg);
 
   if (dwg->header.codepage != 30 && dwg->header.codepage != 29
@@ -1425,10 +1453,13 @@ dxfb_header_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
                 dwg->header.codepage);
     }
 
-#include "header_variables_dxf.spec"
+    // clang-format off
+  #include "header_variables_dxf.spec"
+  // clang-format on
 
   return 0;
 }
+AFL_GCC_POP
 
 static int
 dxfb_classes_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
@@ -1442,8 +1473,8 @@ dxfb_classes_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
       const char *dxfname = dwg->dwg_class[j].dxfname;
       // some classes are now builtin
       if (dat->version >= R_2004
-          && (!strcmp (dxfname, "ACDBPLACEHOLDER")
-              || !strcmp (dxfname, "LAYOUT")))
+          && (strEQc (dxfname, "ACDBPLACEHOLDER")
+              || strEQc (dxfname, "LAYOUT")))
         continue;
       RECORD (CLASS);
       VALUE_TV (dxfname, 1);
@@ -1451,7 +1482,7 @@ dxfb_classes_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
       VALUE_T (dwg->dwg_class[j].appname, 3);
       VALUE_RS (dwg->dwg_class[j].proxyflag, 90);
       SINCE (R_2004) { VALUE_RC (dwg->dwg_class[j].num_instances, 91); }
-      VALUE_RC (dwg->dwg_class[j].wasazombie, 280);
+      VALUE_RC (dwg->dwg_class[j].is_zombie, 280);
       // Is-an-entity. 1f2 for entities, 1f3 for objects
       VALUE_RC (dwg->dwg_class[j].item_class_id == 0x1F2 ? 1 : 0, 281);
     }
@@ -1486,7 +1517,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
           }
         for (i = 0; i < dwg->vport_control.num_entries; i++)
           {
-            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->vports[i]);
+            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             if (obj && obj->type == DWG_TYPE_VPORT)
               {
                 // reordered in the DXF: 2,70,10,11,12,13,14,15,16,...
@@ -1518,7 +1549,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         // here LTYPE_CONTINUOUS is already included
         for (i = 0; i < dwg->ltype_control.num_entries; i++)
           {
-            obj = dwg_ref_object (dwg, _ctrl->linetypes[i]);
+            obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             if (obj && obj->type == DWG_TYPE_LTYPE)
               {
                 error |= dwg_dxfb_LTYPE (dat, obj);
@@ -1537,7 +1568,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         error |= dwg_dxfb_LAYER_CONTROL (dat, ctrl);
         for (i = 0; i < dwg->layer_control.num_entries; i++)
           {
-            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->layers[i]);
+            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             if (obj && obj->type == DWG_TYPE_LAYER)
               {
                 error |= dwg_dxfb_LAYER (dat, obj);
@@ -1558,7 +1589,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         error |= dwg_dxfb_STYLE_CONTROL (dat, ctrl);
         for (i = 0; i < dwg->style_control.num_entries; i++)
           {
-            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->styles[i]);
+            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             if (obj && obj->type == DWG_TYPE_STYLE)
               {
                 error |= dwg_dxfb_STYLE (dat, obj);
@@ -1577,7 +1608,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         error |= dwg_dxfb_VIEW_CONTROL (dat, ctrl);
         for (i = 0; i < dwg->view_control.num_entries; i++)
           {
-            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->views[i]);
+            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             // FIXME implement the other two
             if (obj && obj->type == DWG_TYPE_VIEW)
               error |= dwg_dxfb_VIEW (dat, obj);
@@ -1601,7 +1632,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         error |= dwg_dxfb_UCS_CONTROL (dat, ctrl);
         for (i = 0; i < dwg->ucs_control.num_entries; i++)
           {
-            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->ucs[i]);
+            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             if (obj && obj->type == DWG_TYPE_UCS)
               {
                 error |= dwg_dxfb_UCS (dat, obj);
@@ -1621,7 +1652,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         error |= dwg_dxfb_APPID_CONTROL (dat, ctrl);
         for (i = 0; i < dwg->appid_control.num_entries; i++)
           {
-            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->apps[i]);
+            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             if (obj && obj->type == DWG_TYPE_APPID)
               {
                 error |= dwg_dxfb_APPID (dat, obj);
@@ -1641,7 +1672,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         // ignoring morehandles
         for (i = 0; i < dwg->dimstyle_control.num_entries; i++)
           {
-            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->dimstyles[i]);
+            Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
             if (obj && obj->type == DWG_TYPE_DIMSTYLE)
               {
                 error |= dwg_dxfb_DIMSTYLE (dat, obj);
@@ -1663,8 +1694,7 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
           error |= dwg_dxfb_VPORT_ENTITY_CONTROL (dat, ctrl);
           for (i = 0; i < dwg->vport_entity_control.num_entries; i++)
             {
-              Dwg_Object *obj
-                  = dwg_ref_object (dwg, _ctrl->vport_entity_headers[i]);
+              Dwg_Object *obj = dwg_ref_object (dwg, _ctrl->entries[i]);
               if (obj && obj->type == DWG_TYPE_VPORT_ENTITY_HEADER)
                 {
                   error |= dwg_dxfb_VPORT_ENTITY_HEADER (dat, obj);
@@ -1678,7 +1708,13 @@ dxfb_tables_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
   SINCE (R_13)
   {
     Dwg_Object_BLOCK_CONTROL *_ctrl = &dwg->block_control;
-    Dwg_Object *ctrl = &dwg->object[_ctrl->objid];
+    Dwg_Object *ctrl = _ctrl ? &dwg->object[_ctrl->objid] : NULL;
+    if (!ctrl)
+      {
+        LOG_ERROR ("BLOCK_CONTROL missing");
+        return DWG_ERR_INVALIDDWG;
+      }
+
     // Dwg_Object *obj = dwg_ref_object(dwg, _ctrl->model_space);
     // Dwg_Object *mspace = NULL, *pspace = NULL;
 
@@ -1753,8 +1789,8 @@ dxfb_ENDBLK_empty (Bit_Chain *restrict dat, const Dwg_Object *restrict hdr)
 }
 
 static int
-dxfb_block_write (Bit_Chain *restrict dat, Dwg_Object *restrict hdr,
-                  int *restrict i)
+dxfb_block_write (Bit_Chain *restrict dat, const Dwg_Object *restrict mspace,
+                  Dwg_Object *restrict hdr, int *restrict i)
 {
   int error = 0;
   Dwg_Object *restrict obj = get_first_owned_block (hdr); // BLOCK
@@ -1762,6 +1798,7 @@ dxfb_block_write (Bit_Chain *restrict dat, Dwg_Object *restrict hdr,
       = hdr->tio.object->tio.BLOCK_HEADER;
   Dwg_Object *restrict endblk;
   Dwg_Data *dwg = hdr->parent;
+  unsigned long int mspace_ref = mspace->handle.value;
 
   if (obj)
     error |= dwg_dxfb_object (dat, obj, i);
@@ -1770,16 +1807,20 @@ dxfb_block_write (Bit_Chain *restrict dat, Dwg_Object *restrict hdr,
       LOG_ERROR ("BLOCK_HEADER.block_entity missing");
       return DWG_ERR_INVALIDDWG;
     }
-  // skip *Model_Space UNDERLAY's, they are all under ENTITIES
-  if (hdr == dwg->header_vars.BLOCK_RECORD_MSPACE->obj)
+  // Skip all *Model_Space entities, esp. new ones: UNDERLAY, MULTILEADER, ...
+  // They are all under ENTITIES later.
+  // Note: the objects may vary (e.g. example_2000), but the index not
+  if ((hdr == mspace) || (hdr->index == mspace->index))
     obj = NULL;
   else
     obj = get_first_owned_entity (hdr); // first_entity
   while (obj)
     {
-      if (obj->supertype == DWG_SUPERTYPE_ENTITY)
+      if (obj->supertype == DWG_SUPERTYPE_ENTITY
+          && obj->tio.entity->ownerhandle != NULL
+          && obj->tio.entity->ownerhandle->absolute_ref != mspace_ref)
         error |= dwg_dxfb_object (dat, obj, i);
-      obj = get_next_owned_entity (hdr, obj);
+      obj = get_next_owned_entity (hdr, obj); // until last_entity
     }
   endblk = get_last_owned_block (hdr);
   if (endblk)
@@ -1797,30 +1838,11 @@ static int
 dxfb_blocks_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
   int error = 0;
-  Dwg_Object_BLOCK_CONTROL *_ctrl = &dwg->block_control;
-  // Dwg_Object *ctrl = &dwg->object[_ctrl->objid];
-  /* let's see if this control block is correct... */
-  Dwg_Object_Ref *msref = dwg->header_vars.BLOCK_RECORD_MSPACE;
-  // Dwg_Object_Ref *psref = dwg->header_vars.BLOCK_RECORD_PSPACE;
-  Dwg_Object *mspace;
   int i = 0;
+  Dwg_Object *mspace = dwg_model_space_object (dwg);
 
-  // The modelspace header needs to have an block_entity.
-  // There are cases (r2010 AEC dwgs) where they don't have one.
-  if (msref && msref->obj && msref->obj->type == DWG_TYPE_BLOCK_HEADER
-      && msref->obj->tio.object->tio.BLOCK_HEADER->block_entity)
-    mspace = msref->obj;
-  else
-    mspace = _ctrl->model_space->obj; // these two really should be the same
-
-  // If there's no *Model_Space block skip this BLOCKS section.
-  // Or try handle 1F with r2000+, 17 with r14
-  // obj = get_first_owned_block(hdr);
-  // if (!obj)
-  //  obj = dwg_resolve_handle(dwg, dwg->header.version >= R_2000 ? 0x1f :
-  //  0x17);
-  // if (!obj)
-  //  return 1;
+  if (!mspace)
+    return DWG_ERR_UNHANDLEDCLASS;
 
   SECTION (BLOCKS);
   /* There may be unconnected blocks (not caught by above),
@@ -1833,7 +1855,7 @@ dxfb_blocks_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
         if (dwg->object[i].supertype == DWG_SUPERTYPE_OBJECT
             && dwg->object[i].type == DWG_TYPE_BLOCK_HEADER)
           {
-            error |= dxfb_block_write (dat, &dwg->object[i], &i);
+            error |= dxfb_block_write (dat, mspace, &dwg->object[i], &i);
           }
       }
   }
@@ -1847,7 +1869,9 @@ dxfb_entities_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
   int error = 0;
   int i;
-  Dwg_Object *mspace = dwg->header_vars.BLOCK_RECORD_MSPACE->obj;
+  Dwg_Object *mspace = dwg_model_space_object (dwg);
+  if (!mspace)
+    return DWG_ERR_INVALIDDWG;
 
   SECTION (ENTITIES);
   for (i = 0; (BITCODE_BL)i < dwg->num_objects; i++)
@@ -1884,9 +1908,9 @@ dxfb_objects_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 }
 
 static int
-dxfb_preview_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
+dxfb_thumbnail_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
-  Bit_Chain *pic = (Bit_Chain *)&dwg->picture;
+  Bit_Chain *pic = (Bit_Chain *)&dwg->thumbnail;
   if (pic->chain && pic->size && pic->size > 10)
     {
       SECTION (THUMBNAILIMAGE);
@@ -1900,9 +1924,9 @@ dxfb_preview_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 int
 dwg_write_dxfb (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
-  const int minimal = dwg->opts & 0x10;
+  const int minimal = dwg->opts & DWG_OPTS_MINIMAL;
 
-  loglevel = dwg->opts & 0xf;
+  loglevel = dwg->opts & DWG_OPTS_LOGLEVEL;
   if (dat->from_version == R_INVALID)
     dat->from_version = dat->version;
   fprintf (dat->fh, "AutoCAD Binary DXF\r\n%c%c", 0x1a, 0);
@@ -1942,7 +1966,7 @@ dwg_write_dxfb (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
       }
       SINCE (R_2000)
       {
-        if (dxfb_preview_write (dat, dwg) >= DWG_ERR_CRITICAL)
+        if (dxfb_thumbnail_write (dat, dwg) >= DWG_ERR_CRITICAL)
           goto fail;
       }
     }
